<?php
/**
  * @file sqlite_db.inc
  * @desc データベースアクセス(SQLite用)
  *
  * Ver 1.1 2007/1/1 Y.Kurihara
  * Ver 1.2 2007/7/24 Y.Kurihara 子のSELECT文を使用できるようにした
  * Ver 1.3 2008/5/10 Y.Kurihara エスケープするキーワードを追加
  * Ver 1.4 2007/5/23 Y.Kurihara トランザクション関数追加
  * Ver 2.0 2013/8/16 Y.Kurihara PDO対応版
  * Ver 2.1 2014/6/2  Y.Kurihara Page処理の問題を修正
  * Ver 2.2 2015/12/25  Y.Kurihara limit対応版
  * Ver 2.3 2021/10/27  Y.Kurihara エラーログファイル保存
  *
  * Copyright (c) 2005-2021 Crytus corporation.
  *
 **/
if (!defined("DEBUG")) {
	define("DEBUG", "");
}
/**
 * @name DBConnection
 * @desc データベースアクセス共通処理（PostgreSQL用）
 * sqlite_escape_stringにより、SQLインジェクション対策
 **/
class DBConnection {
    var $_conn = null;	// コネクション
    var $logfile = "";
    /**
	 * @name DBConnection
     * @desc コンストラクタ
     **/
    function __construct() {
		if ($this->logfile == "") {
			$cur = getdate(time());
			$this->logfile = sprintf("log%04d%02d%02d.log", $cur["year"], $cur["mon"], $cur["mday"]);
		}
    }
    /**
     * @name logoutput
	 * @desc ログ出力処理
     */
	function logoutput($level, $message, $file=null, $line=null)
	{
		if (($fp = fopen($this->logfile, "a"))) {
			if ($file && $line) {
				$msg = "File: " . $file . " Line: " . $line . "\r\n";
				fputs($fp, $msg);
			}
			if (is_array($message)) {
				$msg = implode("\r\n", $message);
			} else {
				$msg = $message;
			}
			$cur = getdate(time());
			$buf = sprintf("%02d:%02d %d %s\r\n", $cur["hours"], $cur["minutes"], $loglevel, $msg);
			fputs($fp, $buf);
			fclose($fp);
		}
	}
    /**
     * @name getConnection
	 * @desc データベースアクセス用インスタンス作成
     */
    static function getConnection($DB_URI) {
        static $instance = null;
        if ($instance == null) {
            $instance = new DBConnection();
            // DB 接続
            $ret = $instance->db_open($DB_URI);
			if (!$ret) {
				$instance = null;
			}
        }
        return $instance;
    }
	/**
	 * @name db_open
	 * @desc データベースオープン
	 * @param $uri データベース定義（ホスト、DB名、アカウントなど）
	 * @return 実行結果
	 **/
	function db_open($uri)
	{
		//$this->_conn = sqlite_open($uri);
		$this->_conn = new PDO("sqlite:" . $uri);	// 'sqlite:test1.db'
		if (!$this->_conn) {
			$this->logoutput(1, "sqlite_open error(" . $uri . ")", __FILE__, __LINE__);
			return false;
		}
		return true;
	}
	/**
	 * @name db_exec
	 * @desc SQL実行
	 * @param $sql 実行するSQL文
	 * @return 実行結果
	 **/
	function db_exec($sql, $log=true)
	{
		if (DEBUG && $log) {
			$this->logoutput(0, $sql);		// デバッグ用にSQLをログへ
		}
		$res = $this->_conn->exec($sql);
		if (!$res) {
			$this->logoutput(0, $sql);		// SQLをログへ
			$this->logoutput(1, $this->_conn->errorInfo(), __FILE__, __LINE__);
			var_dump($this->_conn->errorInfo());
			var_dump($sql);
		}
		return $res;
	}
	function db_query($sql, $log=true)
	{
		if (DEBUG && $log) {
			$this->logoutput(0, $sql);		// デバッグ用にSQLをログへ
		}
		$res = $this->_conn->query($sql);
		if (!$res) {
			$this->logoutput(0, $sql);		// SQLをログへ
			$this->logoutput(1, $this->_conn->errorInfo(), __FILE__, __LINE__);
			var_dump($this->_conn->errorInfo());
			var_dump($sql);
		}
		return $res;
	}
	// トランザクション開始
	function db_begin_transaction()
	{
		//return $this->db_exec("BEGIN");
		return false;	// トランザクションなし
	}

	// トランザクション異常終了
	function db_rollback()
	{
		//return $this->db_exec("ROLLBACK");
		return false;	// トランザクションなし
	}

	// トランザクション終了
	function db_commit()
	{
		//return $this->db_exec("COMMIT");
		return false;	// トランザクションなし
	}
	/**
	 * @name escape_string
	 * @desc データベースの文字コード変換とエスケープ
	 * @return 変換後文字列
	 **/
	function escape_string($str, $raw=false)
	{
		global $script_encoding;

		if (is_array($str)) {
			$s = "";
			foreach ($str as $val) {
				if ($s) {
					$s .= ",";
				}
				if ($raw) {
					//$s .= "'" . sqlite_escape_string($val) . "'";
					$s .= $this->_conn->quote($val);
				} else {
					//$s .= "'" . sqlite_escape_string(mb_convert_encoding($val, DB_ENCODING, $script_encoding)) . "'";
					$s .= $this->_conn->quote(mb_convert_encoding($val, DB_ENCODING, $script_encoding));
				}
			}
			return $s;
		}
		if ($raw) {
			return $this->_conn->quote($str);
			//return sqlite_escape_string($str);
		}
		if (!$script_encoding) {
			$script_encoding = "UTF-8";
		}
		return $this->_conn->quote(mb_convert_encoding($str, DB_ENCODING, $script_encoding));
		//return sqlite_escape_string(mb_convert_encoding($str, DB_ENCODING, $script_encoding));
	}
	/**
	 * @name db_close
	 * @desc データベースクローズ
	 * @return なし
	 **/
	function db_close()
	{
		//sqlite_close($this->_conn);
	}
	/**
	 * @name check_field_name
	 * @desc キーワードに一致するフィールド名をエスケープする
	 * @param $name フィールド名
	 * @return なし
	 **/
	function check_field_name($name)
	{
		if (($name == "case")||($name == "type")||($name == "check")) {	// エスケープが必要
			return '`' . $name . '` ';
		}
		return $name;
	}
	/**
	 * @name make_where
	 * @desc 条件の配列からSQLのWHERE文を作成する
	 * @param $where 条件の配列
	 * @return 作成したWHERE文
	 **/
	function make_where($where)
	{
		$count = 0;
		$cond = "";
		if (isset($where)) {
			foreach ($where as $key => $val) {
				$field = $this->check_field_name($val["field"]);
				if ($count > 0) {
					if (isset($val['relation'])) {
						$cond .= " " . $val['relation'] . " ";
					} else {
						$cond .= " and ";
					}
				}
				if (isset($val['nest'])) {
					$cond .= "(" . $this->make_where($val['nest']) . ")";

				} else {
					if (($val['cond'] == 'in')||($val['cond'] == 'not in')) {
						if (isset($val['value'])) {
							$cond .= $field . ' ' . $val['cond'] . " (" . $this->escape_string($val['value']) . ")";
						}
						if (isset($val['select'])) {
							$sql = $this->make_select($val['select']["table"], $val['select']["where"], $val['select']["order"], $val['select']["page"], $val['select']["fields"]);
							$cond .= $field . ' ' . $val['cond'] . " (" . $sql . ")";
						}
					} else if ($val['cond'] == 'any') {	// pgsqlのみ
						$cond .= "'" . $this->escape_string($val['value']) . "'=any(" . $field . ")";
					} else if ($val['cond'] == 'all') {	// pgsqlのみ
						$cond .= "'" . $this->escape_string($val['value']) . "'=all(" . $field . ")";
/*
					} else if ($val['cond'] == 'is') {
						$cond .= $field . ' is ' . $val['value'];
*/
					} else {
						if ($val['cond'] == "") {
							$op = "=";
						} else {
							$op = $val['cond'];
						}
						if (isset($val['expression'])) {
							if (isset($val["field"])) {
								$cond .= $field . ' ' . $op . " " . $this->escape_string($val['expression']);
							} else {
								$cond .= $val['expression'];
							}
						} else {
							if ($val['value'] === "now()") {
							//	$cond .= $field . ' ' . $op . " sysdate()";
								$cond .= $field . ' ' . $op . "'" . date("Y-m-d H:i:s") . "'";
							} else {
								if (is_array($val['value'])) {
									if ($op == "=") {
										$rel = " or ";
									} else {
										$rel = " and ";
									}
									$cond .= "(";
									$n = 0;
									foreach ($val['value'] as $val2) {
										if ($n) {
											$cond .= $rel;
										}
										//$cond .= $field . ' ' . $op . " '" . $this->escape_string($val2) . "'";
										$cond .= $field . ' ' . $op . " " . $this->escape_string($val2) . "";
										$n++;
									}
									$cond .= ")";
								} else {
									//$cond .= $field . ' ' . $op . " '" . $this->escape_string($val['value']) . "'";
									$cond .= $field . ' ' . $op . " " . $this->escape_string($val['value']) . "";
								}
							}
						}
					}
				}
				$count++;
			}
		}
		return $cond;
	}
	/**
	 * @name insert_data
	 * @desc データの挿入
	 * @param $table 挿入するテーブル名
	 * @param $data 挿入するデータ配列（配列のキーが項目名、値がデータ）
	 * @return 実行結果
	 **/
	function insert_data($table, $data, $raw=false)
	{
		$value = '';
		$field = '';
		foreach ($data as $key => $val) {
			if ($value != "") {
				$value .= ",";
				$field .= ",";
			}
			$field .= $this->check_field_name($key);
			if (isset($val)) {
				if (is_array($val)) {
					if (count($val)) {
						$value .= "'{{";
						$cnt = 0;
						foreach ($val as $key2 => $val2) {
							if ($cnt > 0) {
								$value .= ",";
							}
							$value .= '"' . $this->escape_string($val2) . '"';
							$cnt++;
						}
						$value .= "}}'";
					} else {
						$value .= "NULL";
					}
				} else {
					if ($val === 'now()') {	// 現在日時
						//$value .= "sysdate()";
						$value .= "'" . date("Y-m-d H:i:s") . "'";
					} else {
						//$value .= "'" . $this->escape_string($val, $raw) . "'";
						$value .= "" . $this->escape_string($val, $raw) . "";
					}
				}
			} else {
				$value .= "NULL";
			}
		}
		if ($value != "") {
			$sql = "insert into " . $table . " (" . $field . ") values (" . $value . ")";
			if (!$this->db_exec($sql, !$raw)) {
				return false;
			}
		}
		return $this->_conn->lastInsertId();	// ROWIDを返す
		//return sqlite_last_insert_rowid($this->_conn);	// ROWIDを返す
	}
	/**
	 * @name update_data
	 * @desc データの更新
	 * @param $table 更新するテーブル名
	 * @param $data 更新するデータ配列（配列のキーが項目名、値がデータ）
	 * @param $where 更新するデータを特定するWHERE条件
	 * @return 実行結果
	 **/
	function update_data($table, $data, $where, $raw=false)
	{
		$value = '';
		foreach ($data as $key => $val) {
			if ($value != "") {
				$value .= ",";
			}
			$field = $this->check_field_name($key);
			if (is_array($val)) {
				if (count($val)) {
					$value .= $field . "='{{";
					$cnt = 0;
					foreach ($val as $key2 => $val2) {
						if ($cnt > 0) {
							$value .= ",";
						}
						$value .= '"' . $this->escape_string($val2, $raw) . '"';
						$cnt++;
					}
					$value .= "}}'";
				} else {
					$value .= $field . "=NULL";
				}
			} else if ($val === 'now()') {	// 現在日時に更新
			//	$value .= $field . "=sysdate()";
				$value .= $field . "=" . "'" . date("Y-m-d H:i:s") . "'";
			} else if ($val === 'inc()') {	// 値をインクリメント（固有機能）
				$value .= $field . "=" . $val . '+1';
			} else if ($val === '') {
				$value .= $field . "=''";
			} else {
				//$value .= $field . "='" . $this->escape_string($val, $raw) . "'";
				$value .= $field . "=" . $this->escape_string($val, $raw) . "";
			}
		}
		// WHERE作成
		$cond = $this->make_where($where);
		if ($value != "") {
			if ($cond != "") {
				$sql = "update " . $table . " set " . $value . " where " . $cond;
			} else {
				$sql = "update " . $table . " set " . $value;
			}
			return $this->db_exec($sql, !$raw);
		}
		return false;
	}
	/**
	 * @name delete_data
	 * @desc データの削除
	 * @param $table 削除するテーブル名
	 * @param $where 削除するデータを特定するWHERE条件
	 * @return 実行結果
	 **/
	function delete_data($table, $where)
	{
		// WHERE作成
		$cond = $this->make_where($where);
		if ($cond != "") {
			$sql = "delete from " . $table . " where " . $cond;
			return $this->db_exec($sql);
		}
		return false;
	}
	/**
	 * @name count_data
	 * @desc 条件に一致するデータの件数を得る
	 * @param $table カウントするテーブル名
	 * @param $where カウントするデータを特定するWHERE条件
	 * @return 該当データ件数
	 **/
	function count_data($table, $where)
	{
		$data = array('count' => 0);
		$sql = "select count(*) from " . $table;
		$cond = $this->make_where($where);
		if ($cond != "") {
			$sql .= " where " . $cond;
		}
		$ret = $this->db_query($sql);
		if ($ret) {
			$row = $ret->fetch(PDO::FETCH_NUM);
			$data['count'] = $row[0];
		} else {
			$data['count'] = 0;
			$data['error'] = 1;
			//$data['message'] = sqlite_error_string($this->_conn);
			$data['message'] = $this->_conn->errorInfo();
		}
		return $data;
	}

	/**
	 * @name make_select
	 * @desc SELECT文の作成
	 * @param $table 検索するテーブル名
	 * @param $where 検索するデータを特定するWHERE条件
	 * @param $order データの出力順（省略可）
	 * @param $page 出力するデータの位置と件数（省略すると全件）
	 * @param $fields 取得するフィールド名（省略すると全フィールド）
	 * @return sql文
	 **/
	function make_select($table, $where, $order=NULL, $page=NULL, $fields=NULL)
	{
		if ($fields) {
			if (is_array($fields)) {
				$sql = "select " . join(",", $fields) . " from " . $table;
			} else {
				$sql = "select " . $fields . " from " . $table;
			}
		} else {
			$sql = "select * from " . $table;
		}
		$cond = $this->make_where($where);
		if ($cond != "") {
			$sql .= " where " . $cond;
		}
		if (!is_null($order)) {
			$sql .= " order by ";
			$count = 0;
			foreach ($order as $key => $val) {
				if ($count > 0) {
					$sql .= ",";
				}
				$sql .= $val['field'];
				if (isset($val['desc'])) {
					$sql .= " desc";
				}
				$count++;
			}
		}
		if (!is_null($page)) {
			if (!isset($page['offset'])) {
				$page['offset'] = '0';
			}
		//	$sql .= " offset " . $page['offset'] . " limit " .  $page['limit'];
			$sql .= " limit " .  $page['offset'] . "," . $page['limit'];
		}
		return $sql;
	}
	/**
	 * @name search_data
	 * @desc データの検索
	 * @param $table 検索するテーブル名
	 * @param $where 検索するデータを特定するWHERE条件
	 * @param $order データの出力順（省略可）
	 * @param $page 出力するデータの位置と件数（省略すると全件）
	 * @param $fields 取得するフィールド名（省略すると全フィールド）
	 * @return 実行結果
	 **/
	function search_data($table, $where, $order=NULL, $page=NULL, $fields=NULL, $raw=false)
	{
		$sql = $this->make_select($table, $where, $order, $page, $fields);
		return $this->search_sql($sql, $raw, null);
//		return $this->search_sql($sql, $raw, $page);
	}

	/**
	 * @name search_sql
	 * @desc データの検索(SQL直接実行)
	 *	複数テーブルを参照したり、副問い合わせを使用する場合に、直接SQLを実行する。
	 * @param $table 検索するテーブル名
	 * @param $where 検索するデータを特定するWHERE条件
	 * @param $order データの出力順（省略可）
	 * @param $page 出力するデータの位置と件数（省略すると全件）
	 * @return 実行結果
	 **/
	function search_sql($sql, $raw=false, $page=array())
	{
		global $script_encoding;

		$ret = $this->db_query($sql);
		if ($ret) {
			$count = 999999;
			if ($page) {
				$start = $page["offset"];
				if (($count - $start) > $page["limit"]) {
					$count = $page["limit"];
				}
			} else {
				$start = 0;
				$end = $count;
			}
			for ($row = 0; $row < ($start + $count); $row++) {
				$rec = $ret->fetch(PDO::FETCH_ASSOC);
				if (!$rec) break;
				if ($row < $start) continue;
				$data[$row] = $rec;
				if (!$raw) {
					foreach ($data[$row] as $key => $val) {
						$val = mb_convert_encoding($val, $script_encoding, DB_ENCODING);
						// 配列データを展開(実データに「{{～}}」が含まれるとデータが破壊されるので注意)
						if (preg_match('/^{{(.+)}}/', $val, $match)) {
							$ary = explode(",", $match[1]);
							if ($ary) {
								$data[$row][$key] = $ary;
								foreach ($data[$row][$key] as $key2 => $val2) {
									if (preg_match('/"(.+)"/', $val2, $match)) {
										$data[$row][$key][$key2] = $match[1];
									}
								}
							}
						} else {
							$data[$row][$key] = $val;
						}
					}
				}
			}
			$result['count'] = $count;
			$result['data'] = $data;
			return $result;
		} else {
			$data['count'] = 0;
			$data['error'] = 1;
			//$data['message'] = sqlite_error_string($this->_conn);
			$data['message'] = $this->_conn->errorInfo();
		}
		return $data;
	}
}
